/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    IdenticalInstanceFilter.java
 *    Copyright (C) 2007 Reuben Evans   
 */
package weka.filters.unsupervised.instance;

import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;

import weka.core.Capabilities;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.Utils;
import weka.core.Capabilities.Capability;
import weka.filters.SimpleBatchFilter;
import weka.core.Attribute;

/**
 * A Filter to merge or delete Identical Instances 
 * 
 * @author Reuben Evans rjee1 at cs dot waikato dot ac dot nz
 */
public class IdenticalInstanceFilter extends SimpleBatchFilter {

	/**
	 * Wether to merge or delete identical instances
	 */
	private boolean m_Merge = false;
	
	/**
	 * The Verion ID date and time of last breaking Edit 
	 */
	private static final long serialVersionUID = 200703301210L;

	/**
	 * Sets and returns the output format which is the same as the input format
	 * 
	 * @return 				the optput format - an empty copy of the input format
	 * @see			 		weka.filters.SimpleFilter#determineOutputFormat(weka.core.Instances)
	 */
	protected Instances determineOutputFormat(Instances inputFormat) {
	    Instances result = new Instances(inputFormat, 0);
	    return result;
	  }
	  
	  /** 
	   * Returns the Capabilities of this filter. Enables all
	   *
	   * @return            the capabilities of this object
	   * @see               Capabilities
	   */
	  public Capabilities getCapabilities() {
	    Capabilities 	result;

	    result = new Capabilities(this);
	    result.setMinimumNumberInstances(0);
	    result.enableAllAttributes();
	    result.enableAllClasses();
	    result.enable(Capability.NO_CLASS);
	    return result;
	  }
	  
	  /**
	   * Returns a string describing this filter
	   *
	   * @return a description of the filter suitable for
	   * displaying in the explorer/experimenter gui
	   */	
	@Override
	public String globalInfo() {
		return "An Instance filter that ether deletes or merges instances that are identical";
	}

	/* (non-Javadoc)
	 * @see weka.filters.SimpleFilter#process(weka.core.Instances)
	 */
	@Override
	protected Instances process(Instances instances) throws Exception {
		Instances result = new Instances(instances,0);
		boolean notDuplicate = true;
		for(int i = 0;i<instances.numInstances();i++){
			for(int j = 0;j<result.numInstances() && notDuplicate;j++)
				if(equal(instances.instance(i),result.instance(j))){
					notDuplicate = false;
					if(m_Merge)
					result.instance(j).setWeight(result.instance(j).weight()+instances.instance(i).weight());
				}
			if(notDuplicate)
				result.add(instances.instance(i));
			notDuplicate = true;
		}
		return result;
	}
	
	
	
	private boolean equal(Instance first, Instance second){
		if(first.numAttributes() != second.numAttributes())
			return false;
		for(int i = 0;i<first.numAttributes();i++)
			if (first.attribute(i).type()!=second.attribute(i).type()) {
				return false;
			} else if (first.attribute(i).type()==Attribute.NUMERIC) {
				if (Math.abs(first.value(i)-second.value(i))>1E-6) {
					return false;
				}
      } else if (first.attribute(i).type()==Attribute.NOMINAL) {
				if (first.value(i)!=second.value(i)) {
					return false;
				}
			} else {
				return false;
			}

		return true;
	}
	
	/**
	   * Returns an enumeration describing the available options.
	   *
	   * @return an enumeration of all the available options.
	   */
	  public Enumeration listOptions() {
	    Vector<Option> result = new Vector<Option>();

	    result.addElement(new Option(
	              "\tTells the filter to merge rather than delete.",
	              "M", 0, "-M"));
	    
	    Enumeration<Option> e = super.listOptions();
	    while(e.hasMoreElements())
	    	result.add(e.nextElement());
	    
	    return result.elements();
	  }

	  /**
	   * Parses a list of options for this object. 

	   *
	   * @param options the list of options as an array of strings
	   * @throws Exception if an option is not supported
	   */
	  public void setOptions(String[] options) throws Exception { 
	    setMerge(Utils.getFlag('M', options));
	    super.setOptions(options);
	  }

	  /**
	   * Gets the current settings of the filter.
	   *
	   * @return an array of strings suitable for passing to setOptions
	   */
	  public String[] getOptions() {
	    Vector<String> result;

	    result = new Vector<String>();

	    if (getMerge())
	      result.add("-M");
	    
	    result.addAll(Arrays.asList(super.getOptions()));
	    
	    return (String[]) result.toArray(new String[result.size()]);
	  }

	/**
	 * Gets the Value of Merge true if merging flase if deleting
	 * 
	 * @return true if merging instance false if deleting
	 */
	public boolean getMerge() {
		return m_Merge;
	}

	/**
	 * Sets the merge option
	 * 
	 * @param merge wether the filter should merge(true) or delete(false)
	 */
	public void setMerge(boolean merge) {
		this.m_Merge = merge;
	}	

	public static void main(String[] args) {
		runFilter(new IdenticalInstanceFilter(), args);
	}
}
